from . import options
from .semantics import *

NAME = 'arithmetic'
MUTATORS = ['arith-constants', 'arith-negate-relations', 'split-nary-relations', 'strengthen-relations']

def is_arithmetic_relation(node):
    if not has_name(node):
        return False
    return get_name(node) in ['=', '<', '>', '>=', '<=', '!=', '<>']

class ArithmeticSimplifyConstant:
    """Replace a constant by a simpler version (smaller or fewer decimal places)."""
    def filter(self, node):
        return is_arithmetic_constant(node) and float(node) not in [0, 1]
    def mutations(self, node):
        f = float(node)
        if int(f) == f:
            i = int(f)
            return [str(i // 2), str(i // 10)]
        return [str(int(f)), node[:-1]]
    def global_mutations(self, linput, ginput):
        return [ substitute(ginput, {linput: rep}) for rep in self.mutations(linput) ]
    def __str__(self):
        return 'simplify arithmetic constant'

class ArithmeticNegateRelations:
    """Replace a negation around a relation by the inverse relation."""
    def filter(self, node):
        return is_not(node) and is_arithmetic_relation(node[1])
    def mutations(self, node):
        negator = { '<': '>=', '<=': '>', '!=': '=', '<>': '=', '>=': '<', '>': '<=' }
        if node[1][0] in negator:
            return [[negator[node[1][0]]] + node[1][1:]]
        return []
    def __str__(self):
        return 'push negation into relation'

class ArithmeticSplitNaryRelations:
    """Split n-ary relations using transitivity."""
    def filter(self, node):
        return is_arithmetic_relation(node) and len(node) > 3
    def mutations(self, node):
        split = [[get_name(node), node[i], node[i + 1]] for i in range(1, len(node) - 1)]
        return [['and', *split]]
    def __str__(self):
        return 'split n-ary relation'

class ArithmeticStrengthenRelations:
    """Replace a relation by a stronger relation."""
    def filter(self, node):
        return is_arithmetic_relation(node)
    def mutations(self, node):
        negator = { '<': ['='], '>': ['='], '<=': ['<', '='], '>=': ['>', '='] }
        if node[0] in negator:
            return [[rel] + node[1:] for rel in negator[node[0]]]
        return []
    def __str__(self):
        return 'strengthen relation'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'arithmetic mutators')
    options.add_mutator_argument(argparser, 'arith-constants', True, 'replaces constants by simpler ones')
    options.add_mutator_argument(argparser, 'arith-negate-relations', True, 'push negations inside of relations')
    options.add_mutator_argument(argparser, 'arith-split-nary-relations', True, 'split n-ary relations')
    options.add_mutator_argument(argparser, 'arith-strengthen-relations', True, 'strengthen relations')

def collect_mutators(args):
    res = []
    if args.mutator_arithmetic:
        if args.mutator_arith_constants:
            res.append(ArithmeticSimplifyConstant())
        if args.mutator_arith_negate_relations:
            res.append(ArithmeticNegateRelations())
        if args.mutator_arith_split_nary_relations:
            res.append(ArithmeticSplitNaryRelations())
        if args.mutator_arith_strengthen_relations:
            res.append(ArithmeticStrengthenRelations())
    return res



import collections
import copy
import progressbar

from . import mutator_options
from . import semantics

Mutation = collections.namedtuple('Mutation', ['nodeid', 'name', 'localm', 'globalm'])

enabled_mutators = []

def collect_mutators(args):
    """Initializes the list of all active mutators."""
    global enabled_mutators
    enabled_mutators = mutator_options.collect_mutators(args)

class MutationGenerator:
    def __init__(self, skip):
        self.__node_count = 0
        self.__node_skip = skip

    def __mutate_node(self, linput, ginput):
        """Apply all active mutators to the given node.
        Returns a list of all possible mutations as tuples :code:`(name, local, global)`
        where :code:`local` is a modification of the current node and :code:`global` is
        a modification of the whole input and one of those is always :code:`None`."""
        res = []
        for m in enabled_mutators:
            try:
                if hasattr(m, 'filter') and not m.filter(linput):
                    continue
                if hasattr(m, 'mutations'):
                    res = res + list(map(lambda x: Mutation(self.__node_count, str(m), x, None), m.mutations(linput)))
                if hasattr(m, 'global_mutations'):
                    res = res + list(map(lambda x: Mutation(self.__node_count, "(global) " + str(m), None, x), m.global_mutations(linput, ginput)))
            except:
                pass
        return res

    def __generate_mutations(self, linput, ginput, prg):
        """Generate mutations from the given original, updating the progress bar."""
        prg.update(prg.currval + 1)
        if self.__node_count >= self.__node_skip:
            yield from self.__mutate_node(linput, ginput)
        self.__node_count += 1
        if isinstance(linput, list):
            for i, o in enumerate(linput):
                cand = copy.copy(linput)
                for mutated in self.__generate_mutations(o, ginput, prg):
                    if mutated.localm is not None:
                        cand[i] = mutated.localm
                        yield Mutation(mutated.nodeid, mutated.name, cand, mutated.globalm)
                    if mutated.globalm is not None:
                        yield mutated

    def generate_mutations(self, original):
        """A generator that produces all possible mutations from the given original."""
        semantics.collect_information(original)
        s = semantics.node_count(original)
        widgets = [progressbar.Bar(), ' ', progressbar.Counter(), ' / ', str(s)]
        prg = progressbar.ProgressBar(maxval = s, widgets = widgets)
        prg.start()
        prg.update_interval = 1
        for mutated in self.__generate_mutations(original, original, prg):
            if mutated.localm is not None:
                yield (mutated.nodeid, mutated.name, mutated.localm)
            if mutated.globalm is not None:
                yield (mutated.nodeid, mutated.name, mutated.globalm)



import argparse

from . import mutators_arithmetic
from . import mutators_bitvectors
from . import mutators_boolean
from . import mutators_core
from . import mutators_smtlib
from . import mutators_strings

def disable(namespace, option):
    setattr(namespace, 'mutator_{}'.format(option.replace('-', '_')), False)

def disable_all(namespace, options):
    for opt in options:
        disable(namespace, opt)

class AgressiveAction(argparse.Action):
    """Mode that only checks aggressive mutations."""
    def __call__(self, parser, namespace, values, option_string = None):
        setattr(namespace, 'mode_aggressive', True)
        disable(namespace, mutators_arithmetic.NAME)
        disable(namespace, mutators_bitvectors.NAME)
        disable(namespace, mutators_boolean.NAME)
        disable_all(namespace, mutators_core.MUTATORS)
        disable(namespace, mutators_strings.NAME)
        setattr(namespace, 'mutator_constants', True)
        setattr(namespace, 'mutator_erase_children', True)
        setattr(namespace, 'mutator_inline_functions', True)
        setattr(namespace, 'mutator_replace_by_variable', True)
        setattr(namespace, 'mutator_substitute_children', True)

class BeautifyAction(argparse.Action):
    """Mode that enables mutations merely beautify the output."""
    def __call__(self, parser, namespace, values, option_string = None):
        setattr(namespace, 'mutator_variable_names', True)
        setattr(namespace, 'wrap_lines', True)

class LetEliminationAction(argparse.Action):
    """Mode that only checks for let eliminations."""
    def __call__(self, parser, namespace, values, option_string = None):
        setattr(namespace, 'mode_let_elimination', True)
        disable(namespace, mutators_arithmetic.NAME)
        disable(namespace, mutators_bitvectors.NAME)
        disable(namespace, mutators_boolean.NAME)
        disable_all(namespace, mutators_core.MUTATORS)
        disable(namespace, mutators_smtlib.NAME)
        disable(namespace, mutators_strings.NAME)
        setattr(namespace, 'mutator_let_elimination', True)
        setattr(namespace, 'mutator_let_substitution', True)

class ReductionOnlyAction(argparse.Action):
    """Mode that only checks mutations that reduce the number of nodes."""
    def __call__(self, parser, namespace, values, option_string = None):
        setattr(namespace, 'mode_reduction_only', True)
        setattr(namespace, 'mutator_sort_children', False)

class TopLevelOnlyAction(argparse.Action):
    """Mode that only uses top level binary reduction."""
    def __call__(self, parser, namespace, values, option_string = None):
        disable(namespace, mutators_arithmetic.NAME)
        disable(namespace, mutators_bitvectors.NAME)
        disable(namespace, mutators_boolean.NAME)
        disable_all(namespace, mutators_core.MUTATORS)
        disable(namespace, mutators_smtlib.NAME)
        disable(namespace, mutators_strings.NAME)
        setattr(namespace, 'mutator_top_level_binary_reduction', True)

def collect_mutator_modes(argparser):
    argparser.add_argument('--mode-aggressive', default = False, nargs = 0, action = AgressiveAction, help = 'agressively minimize')
    argparser.add_argument('--aggressiveness', metavar = 'perc', type = float, default = 0.01,
                           help = 'percentage of the input a mutators needs to remove')
    argparser.add_argument('--mode-beautify', default = False, nargs = 0, action = BeautifyAction, help = 'enables beautification mutators')
    argparser.add_argument('--mode-let-elimination', default = False, nargs = 0, action = LetEliminationAction, help = 'only eliminate let binders')
    argparser.add_argument('--mode-reduction-only', default = False, nargs = 0, action = ReductionOnlyAction, help = 'only allow reducing mutations')
    argparser.add_argument('--mode-top-level-only', default = False, nargs = 0, action = TopLevelOnlyAction, help = 'use top level binary reduction')

def collect_mutator_options(argparser):
    """Adds all options related to mutators to the given argument parser."""
    mutators_core.collect_mutator_options(argparser.add_argument_group('core mutator arguments'))
    mutators_boolean.collect_mutator_options(argparser.add_argument_group('boolean mutator arguments'))
    mutators_arithmetic.collect_mutator_options(argparser.add_argument_group('arithmetic mutator arguments'))
    mutators_bitvectors.collect_mutator_options(argparser.add_argument_group('bitvector mutator arguments'))
    mutators_smtlib.collect_mutator_options(argparser.add_argument_group('smtlib mutator arguments'))
    mutators_strings.collect_mutator_options(argparser.add_argument_group('strings mutator arguments'))

def collect_mutators(args):
    """Initializes the list of all active mutators."""
    res = []
    res += mutators_core.collect_mutators(args)
    res += mutators_boolean.collect_mutators(args)
    res += mutators_arithmetic.collect_mutators(args)
    res += mutators_bitvectors.collect_mutators(args)
    res += mutators_smtlib.collect_mutators(args)
    res += mutators_strings.collect_mutators(args)
    return res



from . import options
from .semantics import *

NAME = 'bitvector'
MUTATORS = ['bv-constants', 'bv-eval-extract', 'bv-ite-to-bvcomp', 'bv-zero-concat']

def get_bitvector_constant_value(node):
    assert is_bitvector_constant(node)
    if is_leaf(node):
        if node.startswith('#b'):
            return (int(node[2:], 2), len(node[2:]))
        if node.startswith('#x'):
            return (int(node[2:], 16), len(node[2:]) * 4)
        assert False
    return (int(node[1][2:]), node[2])

def possible_bitvector_widths_imp(definition):
    if is_bitvector_type(definition):
        return [definition[2]]
    if not is_leaf(definition):
        return [w for arg in definition for w in possible_bitvector_widths_imp(arg)]
    return []

def possible_bitvector_widths(node):
    if has_type(node):
        assert is_bitvector_type(get_type(node))
        return [get_type(node)[2]]
    widths = set()
    for bvtype in get_variable_info().values():
        for wid in possible_bitvector_widths_imp(bvtype):
            widths.add(wid)
    return list(widths)

class BVConcatToZeroExtend:
    """Replace a :code:`concat` with zero by :code:`zero_extend`."""
    def filter(self, node):
        if not has_name(node) or get_name(node) != 'concat':
            return False
        if not is_bitvector_constant(node[1]):
            return False
        return get_bitvector_constant_value(node[1])[0] == 0
    def mutations(self, node):
        return [
            [['_', 'zero_extend', get_bitvector_constant_value(node[1])[1]], node[2]]
        ]
    def __str__(self):
        return 'replace concat by zero_extend'

class BVExtractConstants:
    """Evaluates a bitvector :code:`extract` if it is applied to a constant."""
    def filter(self, node):
        return is_bitvector_extract(node) and is_bitvector_constant(node[1])
    def mutations(self, node):
        upper = int(node[0][2])
        lower = int(node[0][3])
        constant = get_bitvector_constant_value(node[1])[0]
        constant = constant % (2**(upper + 1))
        constant -= constant % (2**lower)
        return [['_', 'bv{}'.format(constant), str(upper - lower + 1)]]
    def __str__(self):
        return 'evaluate bitvector extract on constant'

class BVOneZeroITE:
    """Replace an :code:`ite` with :code:`bv1`/:code:`bv0` cases by :code:`bvcomp`."""
    def filter(self, node):
        if not is_ite(node):
            return False
        if not has_name(node[1]) or get_name(node[1]) != '=' or len(node[1]) != 3:
            return False
        if not is_bitvector_constant(node[2]) or get_bitvector_constant_value(node[2]) != (1, '1'):
            return False
        if not is_bitvector_constant(node[3]) or get_bitvector_constant_value(node[3]) != (0, '1'):
            return False
        return True
    def mutations(self, node):
        return [['bvcomp', node[1][1], node[1][2]]]
    def __str__(self):
        return 'eliminate ite with bv1 / bv0 cases'

class BVSimplifyConstant:
    """Replace a constant by a simpler version (smaller value)."""
    def filter(self, node):
        return is_bitvector_constant(node) and get_bitvector_constant_value(node)[0] not in [0, 1]
    def mutations(self, node):
        val, width = get_bitvector_constant_value(node)
        return ['#b{{:0>{}b}}'.format(width).format(v) for v in [val // 2, val // 10]]
    def global_mutations(self, linput, ginput):
        return [ substitute_repr(ginput, {repr(linput): rep}) for rep in self.mutations(linput) ]
    def __str__(self):
        return 'simplify bitvector constant'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'bitvector mutators')
    options.add_mutator_argument(argparser, 'bv-constants', True, 'replaces constants by simpler ones')
    options.add_mutator_argument(argparser, 'bv-eval-extract', True, 'evaluate bitvector extract on constants')
    options.add_mutator_argument(argparser, 'bv-ite-to-bvcomp', True, 'replaces bv1/bv0 ites by bvcomp')
    options.add_mutator_argument(argparser, 'bv-zero-concat', True, 'replaces concat with zero by zero_extend')

def collect_mutators(args):
    res = []
    if args.mutator_bitvector:
        if args.mutator_bv_constants:
            res.append(BVSimplifyConstant())
        if args.mutator_bv_eval_extract:
            res.append(BVExtractConstants())
        if args.mutator_bv_ite_to_bvcomp:
            res.append(BVOneZeroITE())
        if args.mutator_bv_zero_concat:
            res.append(BVConcatToZeroExtend())
    return res



import os

def version_from_git(gitdir):
    import re
    import subprocess
    GIT_VERSION = subprocess.check_output(['git', 'describe', '--tags'], cwd = gitdir).decode('utf8').strip()
    if re.match('^v[0-9.]+$', GIT_VERSION) is not None:
        return GIT_VERSION[1:]
    else:
        m = re.match('v([0-9]+)\\.([0-9]+)-([0-9]+)-g[a-z0-9]+', GIT_VERSION)
        if m is not None:
            return '{}.{}.dev{}'.format(m.group(1), int(m.group(2)) + 1, m.group(3))
    return GIT_VERSION

def version_from_package_metadata():
    try:
        from importlib import metadata
    except ImportError:
        import importlib_metadata as metadata
    return metadata.version('pydelta-smt')

__git_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '.git')
if os.path.isdir(__git_dir):
    VERSION = version_from_git(__git_dir)
else:
    VERSION = version_from_package_metadata()



import re

from . import options
from .semantics import *

NAME = 'smtlib'
MUTATORS = [
    'check-sat-assuming', 'eliminate-distinct', 'inline-functions', 'let-elimination',
    'let-substitution', 'push-pop-removal', 'simplify-logic', 'simplify-quoted-symbol', 'variable-names'
]

class CheckSatAssuming:
    """Replaces a :code:`check-sat-assuming` by a regular :code:`check-sat`."""
    def filter(self, node):
        return has_name(node) and get_name(node) == 'check-sat-assuming'
    def mutations(self, node):
        return [['check-sat']]
    def __str__(self):
        return 'substitute check-sat-assuming by check-sat'

class EliminateDistinct:
    """Replaces distinct by a negated equality."""
    def filter(self, node):
        return has_name(node) and get_name(node) == 'distinct'
    def mutations(self, node):
        return [['not', ['='] + node[1:]]]
    def __str__(self):
        return 'eliminate distinct'

class InlineDefinedFuns:
    """Explicitly inlines a defined function."""
    def filter(self, node):
        return is_defined_function(node)
    def mutations(self, node):
        return [ get_defined_function(node) ]
    def __str__(self):
        return 'inline defined function'

class LetElimination:
    """Substitutes a :code:`let` expression with its body."""
    def filter(self, node):
        return is_let(node)
    def mutations(self, node):
        return [node[2]]
    def __str__(self):
        return 'eliminate let binder'

class LetSubstitution:
    """Substitutes a variable bound by a :code:`let` binder into the nested term."""
    def filter(self, node):
        return is_let(node)
    def mutations(self, node):
        res = []
        for var in node[1]:
            if contains(node[2], var[0]):
                subs = substitute(node[2], {var[0]: var[1]})
                res.append([node[0], node[1], subs])
        return res
    def __str__(self):
        return 'substitute variable into let body'

class PushPopRemoval:
    """Removes matching :code:`(push)(pop)` pairs. First tries successive pairs, distant ones later."""
    def filter(self, node):
        return not has_name(node)
    def mutations(self, node):
        res = []
        pairs = []
        # identify (push) / (pop) pairs
        stack = []
        for i in range(len(node)):
            if node[i] == ['push']:
                stack.append(i)
            if node[i] == ['pop'] and stack != []:
                pairs.append((stack[-1], i))
                stack.pop()
        # remove directly successive pairs
        for p in pairs:
            if p[0]+1 == p[1]:
                i = p[0]
                res.append(node[:i] + node[i+2:])
        if res != []:
            return res
        # remove non-successive pairs
        for p in pairs:
            r = node[:p[0]] + node[p[0]+1:p[1]] + node[p[1]+1:]
            res.append(r)
        return res
    def __str__(self):
        return 'remove push-pop pair'
class SimplifyLogic:
    """Replaces the logic specified in :code:`(check-logic ...)` by a simpler one."""
    def filter(self, node):
        return has_name(node) and get_name(node) == 'set-logic'
    def mutations(self, node):
        logic = node[1]
        cands = []
        repls = { 'BV': '', 'FP': '', 'UF': '', 'S': '', 'T': '', 'NRA': 'LRA' }
        for r in repls:
            if r in logic:
                cands.append(logic.replace(r, repls[r]))
        return [['set-logic', c] for c in cands]
    def __str__(self):
        return 'simplify logic'

class SimplifyQuotedSymbol:
    """Turns a quoted symbol into a simple symbol."""
    def filter(self, node):
        return is_quoted_symbol(node) and re.match('|[a-zA-Z0-9~!@$%^&*_+=<>.?/-]+|', node) is not None
    def global_mutations(self, linput, ginput):
        return [ substitute(ginput, { linput: get_quoted_symbol(linput) }) ]
    def __str__(self):
        return 'simplify quoted symbol'

class VariableNames:
    """Simplify variable names."""
    def filter(self, node):
        return has_name(node) and get_name(node) == 'declare-fun'
    def global_mutations(self, linput, ginput):
        name = linput[1]
        repl = lambda s: {linput[1]: s}
        if is_quoted_symbol(name):
            name = get_quoted_symbol(linput[1])
            repl = lambda s: {linput[1]: '|' + s + '|'}
        return [
            substitute(ginput, repl(s)) for s in self.__simpler(name)
        ]
    def __simpler(self, name):
        if len(name) < 2:
            return []
        half = len(name) // 2
        return [name[:half], name[:-1]]
    def __str__(self):
        return 'simplify variable name'

def collect_mutator_options(argparser):
    options.add_mutator_argument(argparser, NAME, True, 'smtlib mutators')
    options.add_mutator_argument(argparser, 'check-sat-assuming', True, 'replace check-sat-assuming by check-sat')
    options.add_mutator_argument(argparser, 'eliminate-distinct', True, 'eliminate distinct by negated equalities')
    options.add_mutator_argument(argparser, 'inline-functions', True, 'inline defined functions')
    options.add_mutator_argument(argparser, 'let-elimination', True, 'eliminate let bindings')
    options.add_mutator_argument(argparser, 'let-substitution', True, 'substitute bound variables in let bindings')
    options.add_mutator_argument(argparser, 'push-pop-removal', True, 'remove push-pop pairs')
    options.add_mutator_argument(argparser, 'simplify-logic', True, 'simplify declared logic')
    options.add_mutator_argument(argparser, 'simplify-quoted-symbol', False, 'simplify quoted symbols')
    options.add_mutator_argument(argparser, 'variable-names', False, 'simplify variable names')

def collect_mutators(args):
    res = []
    if args.mutator_smtlib:
        if args.mutator_check_sat_assuming:
            res.append(CheckSatAssuming())
        if args.mutator_eliminate_distinct:
            res.append(EliminateDistinct())
        if args.mutator_inline_functions:
            res.append(InlineDefinedFuns())
        if args.mutator_let_elimination:
            res.append(LetElimination())
        if args.mutator_let_substitution:
            res.append(LetSubstitution())
        if args.mutator_push_pop_removal:
            res.append(PushPopRemoval())
        if args.mutator_simplify_logic:
            res.append(SimplifyLogic())
        if args.mutator_simplify_quoted_symbol:
            res.append(SimplifyQuotedSymbol())
        if args.mutator_variable_names:
            res.append(VariableNames())
    return res



import collections
import logging
import re
import resource
import subprocess
import time

from . import options

ExecResult = collections.namedtuple('ExecResult', ['exitcode', 'stdout', 'stderr', 'runtime'])
__REFERENCE = ExecResult(-1, '', '', -1)
CHECKS = 0
TIMEOUTS = 0

def limit_memory():
    """Apply memory limit given by :code:`--memout`."""
    if options.args().memout != 0:
        resource.setrlimit(resource.RLIMIT_AS, (options.args().memout * 1024 * 1024, resource.RLIM_INFINITY))

def execute(cmd, inputfile):
    """Executes :code:`cmd` on :code:`inputfile`."""
    try:
        global CHECKS
        CHECKS += 1
        if options.args().run_unchecked:
            return ExecResult(0, '', '', 0)
        start = time.time()
        timeout = None if options.args().timeout == 0 else options.args().timeout
        proc = subprocess.Popen(cmd + [inputfile], stdout = subprocess.PIPE, stderr = subprocess.PIPE, preexec_fn = limit_memory)
        out, err = proc.communicate(timeout = timeout)
        duration = time.time() - start
        return ExecResult(proc.returncode, out.decode('utf8').strip(), err.decode('utf8').strip(), duration)
    except subprocess.TimeoutExpired:
        global TIMEOUTS
        TIMEOUTS += 1
        proc.terminate()
        try:
            proc.wait(timeout = 2)
        except subprocess.TimeoutExpired:
            proc.kill()
        try:
            proc.wait(timeout = 2)
        except subprocess.TimeoutExpired:
            logging.warning('Killing pid %d failed. Please check manually to avoid memory exhaustion.', proc.pid)
        return ExecResult(-1, '', '', 0)

def compute_reference(cmd, inputfile):
    """Computes the reference result on the original input.
    Sets an automatic timeout if none was specified.
    """
    global __REFERENCE
    __REFERENCE = execute(cmd, inputfile)
    logging.info('Reference result: exit code %d after %.2f seconds', __REFERENCE.exitcode, __REFERENCE.runtime)
    if options.args().ignore_output:
        logging.info('Reference output is being ignored')
    else:
        logging.info('Reference stdout: \"%s\"', __REFERENCE.stdout)
        logging.info('Reference stderr: \"%s\"', __REFERENCE.stderr)
        if options.args().match_out is not None:
            if not re.search(options.args().match_out, __REFERENCE.stdout):
                logging.error('The pattern for stdout does not match the reference output')
                return False
        if options.args().match_err is not None:
            if not re.search(options.args().match_err, __REFERENCE.stderr):
                logging.error('The pattern for stderr does not match the reference output')
                return False
    if options.args().timeout == 0:
        options.args().timeout = max(int(__REFERENCE.runtime + 1) * 2, 1)
        logging.info('Using automatic timeout of %d seconds (reference run took %.2f seconds)', options.args().timeout, __REFERENCE.runtime)
    return True

def matches_reference(result):
    """Checkes whether the :code:`result` matches the reference result."""
    if options.args().run_unchecked:
        return True
    if not options.args().ignore_exitcode:
        if __REFERENCE.exitcode != result.exitcode:
            return False
    if not options.args().ignore_output:
        if options.args().match_err is None and options.args().match_out is None:
            # if neither --match-err or --match-out are given
            if __REFERENCE.stdout != result.stdout:
                return False
            if __REFERENCE.stderr != result.stderr:
                return False
        else:
            # --match-err or --match-out are given
            if options.args().match_err is not None:
                if not re.search(options.args().match_err, result.stderr):
                    return False
            if options.args().match_out is not None:
                if not re.search(options.args().match_out, result.stdout):
                    return False
    return True



import re

__defined_functions = {}
__defined_variables = {}

def is_leaf(node):
    """Checks whether the :code:`node` is a leaf node."""
    return not isinstance(node, list)

def is_empty(node):
    """Checks whether the :code:`node` is empty."""
    return node == []

def has_name(node):
    """Checks whether the :code:`node` has a name,
    that is its first child is a leaf node."""
    return not is_leaf(node) and not is_empty(node) and is_leaf(node[0])

def get_name(node):
    """Gets the name of the :code:`node`,
    asserting that :code:`has_name(node)`."""
    assert has_name(node)
    return node[0]

def is_quoted_symbol(node):
    """Checks whether the :code:`node` is a quoted symbol."""
    return is_leaf(node) and node[0] == '|' and node[-1] == '|'

def get_quoted_symbol(node):
    """Returns the actual symbol name from a quoted symbol :code:`node`."""
    assert is_quoted_symbol(node)
    return node[1:-1]


def is_smaller(n, m):
    if isinstance(n, str):
        if isinstance(m, str):
            return n < m
        else:
            return True
    else:
        if isinstance(m, str):
            return False
        else:
            return n < m

def is_ite(node):
    """Checks whether :code:`node` is an if-then-else expression."""
    return has_name(node) and get_name(node) == 'ite'

def is_let(node):
    """Checks whether :code:`node` is a let binder."""
    return has_name(node) and get_name(node) == 'let'

def is_defined_function(node):
    """Checks whether :code:`node` is a defined function."""
    if is_leaf(node):
        return node in __defined_functions
    return has_name(node) and get_name(node) in __defined_functions

def get_defined_function(node):
    """Returns the defined function :code:`node`, instantiated with the arguments of :code:`node` if necessary.
    Assumes :code:`__is_defined_functions(node)`."""
    assert is_defined_function(node)
    if is_leaf(node):
        return __defined_functions[node]([])
    return __defined_functions[get_name(node)](node[1:])

def is_nary(node):
    """Checks whether the :code:`node` is a n-ary operator."""
    if is_leaf(node) or not has_name(node):
        return False
    return get_name(node) in [
        '=>', 'and', 'or', 'xor', '=', 'distinct',
        '+', '-', '*', 'div', '/',
        '<=', '<', '>=', '>',
        'bvand', 'bvor', 'bvadd', 'bvmul'
    ]

def node_count(exprs):
    """Counts the number of nodes within :code:`exprs`."""
    if not is_leaf(exprs):
        return 1 + sum(map(node_count, exprs))
    return 1

def iterate_nodes(expr):
    """A generator that performs an pre-order iteration over all nodes."""
    yield expr
    if not is_leaf(expr):
        for e in expr:
            yield from iterate_nodes(e)

def contains(node, sub):
    """Checks whether :code:`node` contains the node :code:`sub`."""
    if node == sub:
        return True
    if is_leaf(node):
        return False
    return any(map(lambda n: contains(n, sub), node))

def substitute(node, repl):
    """Performs substitution recursively within :code:`node`.
    :code:`repl` specifies the substitutions as a dictionary where keys can only be leaf nodes."""
    if is_leaf(node):
        return repl.get(node, node)
    return list(map(lambda n: substitute(n, repl), node))

def substitute_repr(node, repl):
    """Performs substitution recursively within :code:`node`.
    :code:`repl` specifies the substitutions as a dictionary.
    To support substitution of nodes that are not leaf nodes, keys should be :code:`repr(node)`."""
    r = repl.get(repr(node), None)
    if r is not None:
        return r
    if is_leaf(node):
        return node
    return list(map(lambda n: substitute_repr(n, repl), node))

def is_not(node):
    """Checks whether :code:`node` is a negation."""
    return has_name(node) and get_name(node) == 'not'

def is_bitvector_type(node):
    if is_leaf(node) or len(node) != 3:
        return False
    if not has_name(node) or get_name(node) != '_':
        return False
    return node[1] == 'BitVec'

def is_set_type(node):
    if is_leaf(node) or len(node) != 2:
        return False
    if not has_name(node) or get_name(node) != 'Set':
        return False
    return True

def is_boolean_constant(node):
    """Checks whether the :code:`node` is a Boolean constant."""
    return is_leaf(node) and node in ['false', 'true']

def is_arithmetic_constant(node):
    """Checks whether the :code:`node` is an arithmetic constant."""
    return is_leaf(node) and re.match('[0-9]+(\\.[0-9]*)?', node) is not None

def is_int_constant(node):
    """Checks whether the :code:`node` is an int constant."""
    return is_leaf(node) and re.match('^[0-9]+$', node) is not None

def is_real_constant(node):
    """Checks whether the :code:`node` is a real constant."""
    return is_leaf(node) and re.match('^[0-9]+(\\.[0-9]*)?$', node) is not None

def is_string_constant(node):
    """Checks whether the :code:`node` is a string constant."""
    return is_leaf(node) and re.match('^\"[^\"]*\"$', node) is not None

def is_bitvector_constant(node):
    if is_leaf(node):
        if node.startswith('#b'):
            return True
        if node.startswith('#x'):
            return True
        return False
    if len(node) != 3:
        return False
    if not has_name(node) or get_name(node) != '_':
        return False
    return node[1].startswith('bv')

def is_constant(node):
    return is_boolean_constant(node) or is_arithmetic_constant(node) or is_int_constant(node) or is_real_constant(node) or is_string_constant(node) or is_bitvector_constant(node)

def is_indexed_op(node):
    if is_leaf(node) or len(node) < 2:
        return False
    if has_name(node) or not has_name(node[0]):
        return False
    if get_name(node[0]) != '_':
        return False
    return True

def get_indexed_op_name(node):
    assert not is_leaf(node) and not is_leaf(node[0]) and node[0][0] == '_'
    return node[0][1]

def is_bitvector_extend(node):
    if not is_indexed_op(node):
        return False
    return len(node[0]) == 3 and node[0][1] in ['zero_extend', 'sign_extend']

def is_bitvector_extract(node):
    if not is_indexed_op(node):
        return False
    return len(node[0]) == 4 and node[0][1] == 'extract'

def is_bitvector_repeat(node):
    if not is_indexed_op(node):
        return False
    return len(node[0]) == 3 and node[0][1] == 'repeat'

def is_bitvector_rotate(node):
    if not is_indexed_op(node):
        return False
    return len(node[0]) == 3 and node[0][1] in ['rotate_left', 'rotate_right']

def get_bitvector_width(node):
    if is_bitvector_constant(node):
        if is_leaf(node):
            if node.startswith('#b'):
                return len(node[2:])
            if node.startswith('#x'):
                return len(node[2:]) * 4
        return int(node[2])
    if has_type(node):
        assert is_bitvector_type(get_type(node))
        return int(get_type(node)[2])
    if has_name(node):
        if get_name(node) in [
                'bvnot', 'bvand', 'bvor',
                'bvneg', 'bvadd', 'bvmul', 'bvudiv', 'bvurem',
                'bvshl', 'bvshr',
                'bvnand', 'bvnor', 'bvxor', 'bvsub', 'bvsdiv',
                'bvsrem', 'bvsmod', 'bvashr'
        ]:
            return get_bitvector_width(node[1])
        if get_name(node) == 'concat':
            assert len(node) == 3
            return get_bitvector_width(node[1]) + get_bitvector_width(node[2])
        if get_name(node) == 'bvcomp':
            return 1
        if is_bitvector_extend(node):
            return int(node[0][2]) + get_bitvector_width(node[1])
        if is_bitvector_extract(node):
            return int(node[0][2]) - int(node[0][3]) + 1
        if is_bitvector_repeat(node):
            return int(node[0][2]) * get_bitvector_width(node[1])
        if is_bitvector_rotate(node):
            return get_bitvector_width(node[1])
    return -1

def get_constants(const_type):
    """Returns a list of constants for the given type."""
    if const_type == 'Bool':
        return ['false', 'true']
    if const_type == 'Int':
        return ['0', '1']
    if const_type == 'Real':
        return ['0.0', '1.0']
    if is_bitvector_type(const_type):
        return [['_', c, const_type[2]] for c in ['bv0', 'bv1']]
    if is_set_type(const_type):
        return [['as', 'emptyset', const_type]] + [
            ['singleton', c] for c in get_constants(const_type[1])
        ]
    return []

def get_return_type(node):
    """Tries to figure out the return type of the given node.
    Returns :code:`None` if it can not be inferred."""
    if has_type(node):
        return get_type(node)
    if is_boolean_constant(node):
        return 'Bool'
    if is_bitvector_constant(node):
        return ['_', 'BitVec', str(get_bitvector_width(node))]
    if is_int_constant(node):
        return 'Int'
    if is_real_constant(node):
        return 'Real'
    bvwidth = get_bitvector_width(node)
    if bvwidth != -1:
        return ['_', 'BitVec', str(bvwidth)]
    if has_name(node):
        if is_ite(node):
            return get_return_type(node[1])
        # stuff that returns Bool
        if get_name(node) in [
                # core theory
                'not', '=>', 'and', 'or', 'xor', '=', 'distinct',
                # bv theory
                'bvult',
                # fp theory
                'fp.leq', 'fp.lt', 'fp.geq', 'fp.gt', 'fp.eq',
                'fp.isNormal', 'fp.isSubnormal', 'fp.isZero',
                'fp.isInfinite', 'fp.isNaN',
                'fp.isNegative', 'fp.isPositive',
                # int / real theory
                '<=', '<', '>>', '>', 'is_int',
                # sets theory
                'member', 'subset',
                # string theory
                'str.<', 'str.in_re', 'str.<=',
                'str.prefixof', 'str.suffixof', 'str.contains',
                'str.is_digit',
        ]:
            return 'Bool'
        # int theory
        if get_name(node) == '_' and len(node) == 3 and node[1] == 'divisible':
            return 'Bool'
        # stuff that returns Int
        if get_name(node) in [
                'div', 'mod', 'abs', 'to_int',
                # string theory
                'str.len', 'str.indexof', 'str.to_code', 'str.to_int',
                # sets theory
                'card'
        ]:
            return 'Int'
        # stuff that returns Real
        if get_name(node) in ['/', 'to_real', 'fp.to_real']:
            return 'Real'
        if get_name(node) in ['+', '-', '*']:
            if any(map(lambda n: get_return_type(n) == 'Real', node[1:])):
                return 'Real'
            else:
                return 'Int'
    return None

def get_variable_info():
    """Returns the type lookup table."""
    return __defined_variables

def has_type(node):
    """Checks whether :code:`node` was defined to have a certain type.
    Mostly applies to variables."""
    return is_leaf(node) and node in __defined_variables

def get_type(node):
    """Returns the type of :code:`node` if it was defined to have a certain type.
    Assumes :code:`has_type(node)`."""
    assert has_type(node)
    return __defined_variables[node]

def get_variables_with_type(var_type):
    """Returns all variables with the type :code:`var_type`."""
    return [
        v for v in __defined_variables
        if __defined_variables[v] == var_type
    ]


def collect_information(exprs):
    """Initialize global lookups: defined functions and types."""
    global __defined_functions
    global __defined_variables
    __defined_functions = {}
    __defined_variables = {}

    for node in iterate_nodes(exprs):
        if not has_name(node):
            continue
        if get_name(node) == 'declare-const':
            if not len(node) == 3:
                continue
            assert is_leaf(node[1])
            __defined_variables[node[1]] = node[2]
        if get_name(node) == 'declare-fun':
            if not len(node) == 4:
                continue
            assert is_leaf(node[1])
            __defined_variables[node[1]] = node[3]
        if get_name(node) == 'define-fun':
            if not len(node) == 5:
                continue
            assert is_leaf(node[1])
            assert not is_leaf(node[2])
            if not is_leaf(node[3]):
                continue
            __defined_functions[node[1]] = lambda args, node=node: substitute(
                node[4],
                {node[2][i][0]: args[i] for i in range(len(args))}
            )
